# Copyright (c) 2018-2023 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import inspect
import os
import sys
import logging


try:
    import blackboxprotobuf
except ModuleNotFoundError:
    # two abspath because dirname gives an empty string if we run just bbpb.py
    _BASE_DIR = os.path.abspath(
        os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        + "/.."
    )
    sys.path.insert(0, _BASE_DIR + "/lib/")
    sys.path.insert(0, _BASE_DIR + "/mitmproxy/")
    sys.path.insert(0, _BASE_DIR + "/mitmproxy/deps/six/")
    import blackboxprotobuf

from blackboxprotobuf.lib import payloads
from blackboxprotobuf.lib.exceptions import BlackboxProtobufException

import json
from collections.abc import Sequence

from typing import TypeVar, Callable, Optional

from mitmproxy import (
    command,
    contentviews,
    ctx,
    http,
    flow,
    types,
    exceptions,
    websocket,
)
from mitmproxy.tools.console import overlay, signals, keymap


class BlackboxProtobufAddon:
    def __init__(self):
        self.view = BlackboxProtobufView(self)

        self.bbpb_config = blackboxprotobuf.lib.config.Config()
        self.typedef_lookup = {}
        self.project_file = None

    def load(self, loader):
        contentviews.add(self.view)

        loader.add_option(
            "bbpb_project_file",
            typespec=str,
            default="",
            help="Persist known_types and typedef mappings to a project file. File will be written to automatically, so consider keeping a semi-regular backup",
        )

    def _load_project_file(self, project_file: str | None = None):
        self.project_file = ctx.options.bbpb_project_file

        if not project_file and not self.project_file:
            return

        if not project_file:
            project_file = self.project_file

        if not self.project_file:
            return

        logging.info("Loading project data from file")
        try:
            with open(project_file, "r") as f:
                project_data = json.load(f)

            # We update the existing typedef_lookup and bbpb_confg in case we have existing data
            self.typedef_lookup.update(project_data["typedef_lookup"])
            self.bbpb_config.known_types.update(project_data["known_types"])
        except FileNotFoundError:
            # We haven't written anything, so the file might not exist yet
            pass
        self._refresh_view()

    def _save_project_file(self, project_file: str | None = None):
        if not project_file and not self.project_file:
            return

        if not project_file:
            project_file = self.project_file

        logging.info("Writing project data to file")
        data = {
            "typedef_lookup": self.typedef_lookup,
            "known_types": self.bbpb_config.known_types,
        }
        with open(project_file, "w") as f:
            json.dump(data, f, indent=2)

    def configure(self, updates: set[str]):
        if "bbpb_project_file" in updates:
            self._load_project_file()
            self._save_project_file()

    def done(self):
        contentviews.remove(self.view)

    @command.command("bbpb.edit")
    @command.argument("flow_part", type=types.Choice("bbpb.options.edit_part"))
    def bbpb_edit(self, flow_part: str) -> None:
        flow = ctx.master.view.focus.flow

        if flow_part == "request-body":
            message = flow.request
        elif flow_part == "response-body":
            if not flow.response:
                raise exceptions.CommandError(
                    f"Flow part is response-body, but flow has no response"
                )
            message = flow.response
        elif flow_part == "websocket":
            # Edit the last websocket
            message = flow.websocket.messages[-1]
        else:
            raise exceptions.CommandError(f"Got unknown flow_part: {flow_part}")

        message_hash = _message_hash(message.content, message, flow)

        typedef = self.typedef_lookup.get(message_hash)
        message_json, typedef_out, encoding_alg = _decode_protobuf(
            message.content, typedef, self.bbpb_config
        )

        def message_callback(edited_json: str) -> None:
            protobuf_data = blackboxprotobuf.protobuf_from_json(
                edited_json, typedef_out
            )
            data = payloads.encode_payload(protobuf_data, encoding_alg)
            message.content = bytes(data)
            self._refresh_view()

        _spawn_validating_editor(message_json, message_callback)

    @command.command("bbpb.edit_type")
    @command.argument("flow_part", type=types.Choice("bbpb.options.edit_type_part"))
    def bbpb_edit_type(self, flow_part: str) -> None:
        flow = ctx.master.view.focus.flow

        typedef, message_hash = self._resolve_type(flow_part)

        typedef_json = json.dumps(typedef, indent=2)

        def typedef_callback(typedef_json: str) -> None:
            new_typedef = json.loads(typedef_json)
            blackboxprotobuf.validate_typedef(
                new_typedef, typedef
            )  # Validate against old typedef

            blackboxprotobuf.lib.api._strip_typedef_annotations(new_typedef)
            known_type = self.typedef_lookup.get(message_hash)
            if isinstance(known_type, str):
                # This is a named typedef, edit the known typedef instead of the saved value
                self.bbpb_config.known_types[known_type] = new_typedef
                self._save_project_file()
            else:
                # Trusting validate_typedef and not going to try to use the typedef to decode again or reencode
                self.typedef_lookup[message_hash] = new_typedef
                self._save_project_file()

            self._refresh_view()

        _spawn_validating_editor(typedef_json, typedef_callback)

    @command.command("bbpb.apply_type")
    @command.argument("flow_part", type=types.Choice("bbpb.options.edit_type_part"))
    @command.argument("typename", type=types.Choice("bbpb.options.known_types"))
    def bbpb_apply_type(self, flow_part: str, typename: str) -> None:
        flow = ctx.master.view.focus.flow
        if typename not in self.bbpb_config.known_types and typename != "(clear)":
            raise exceptions.CommandError(f"Type {typename} is not a know type")
        flow = ctx.master.view.focus.flow
        if not flow:
            raise exceptions.CommandError("No flow selected.")
        if flow_part.startswith("request") or flow_part.startswith("response"):
            if flow_part == "request-body":
                message = flow.request
            elif flow_part == "response-body":
                if not flow.response:
                    raise exceptions.CommandError(
                        f"Flow part is response-body, but flow has no response"
                    )
                message = flow.response
            message_hash = _message_hash(message.content, message, flow)
            if typename == "(clear)":
                logging.info("popping message hash")
                self.typedef_lookup.pop(message_hash, None)
                self._save_project_file()
                self._refresh_view()
                return

            # Validate that we can decode the message with our new type
            try:
                _decode_protobuf(
                    message.content, typename, self.bbpb_config, fallback=False
                )
            except BlackboxProtobufException as ex:
                raise exceptions.CommandError(
                    f"Error applying type name {typename} to part {flow_part}: {ex}"
                )

        elif flow_part.startswith("websocket"):
            # Websockets don't have a single typedef to edit
            # Instead, we are going to build a typedef based on all the messages
            if flow_part == "websocket-request":
                if not flow.websocket:
                    raise exceptions.CommandError(
                        f"Flow part is websocket-request, but flow is not a websocket"
                    )
                messages = [
                    message
                    for message in flow.websocket.messages
                    if message.from_client
                ]
            elif flow_part == "websocket-response":
                if not flow.websocket:
                    raise exceptions.CommandError(
                        f"Flow part is websocket-response, but flow is not a websocket"
                    )
                messages = [
                    message
                    for message in flow.websocket.messages
                    if not message.from_client
                ]
            if not messages:
                raise exceptions.CommandError(
                    f"Could not find any messages for flow part: {flow_part}"
                )
            message_hash = _message_hash(messages[0].content, messages[0], flow)
            if typename == "(clear)":
                logging.info("popping message hash")
                self.typedef_lookup.pop(message_hash, None)
                self._save_project_file()
                self._refresh_view()
                return
            # Validate that we can decode all messages with our type
            for message in messages:
                try:
                    _decode_protobuf(
                        message.content, typename, self.bbpb_config, fallback=False
                    )
                except BlackboxProtobufException as ex:
                    raise exceptions.CommandError(
                        f"Error applying type name {typename} to part {flow_part}: {ex}"
                    )
        # Success
        self.typedef_lookup[message_hash] = typename
        self._save_project_file()
        self._refresh_view()

    @command.command("bbpb.new_type")
    @command.argument("flow_part", type=types.Choice("bbpb.options.edit_type_part"))
    @command.argument("typename", type=str)
    def bbpb_new_type(self, flow_part: str, typename: str) -> None:
        if typename == "(clear)":
            raise exceptions.CommandError(f"Error: Typename {typename} is not valid.")
        typedef, message_hash = self._resolve_type(flow_part)

        blackboxprotobuf.lib.api._strip_typedef_annotations(typedef)
        self.typedef_lookup[message_hash] = typename
        self.bbpb_config.known_types[typename] = typedef
        self._save_project_file()

        self._refresh_view()

    def _resolve_type(self, flow_part):
        flow = ctx.master.view.focus.flow
        if not flow:
            raise exceptions.CommandError("No flow selected.")
        if flow_part.startswith("request") or flow_part.startswith("response"):
            if flow_part == "request-body":
                message = flow.request
            elif flow_part == "response-body":
                if not flow.response:
                    raise exceptions.CommandError(
                        f"Flow part is response-body, but flow has no response"
                    )
                message = flow.response
            message_hash = _message_hash(message.content, message, flow)
            saved_typedef = self.typedef_lookup.get(message_hash)
            message_json, typedef, encoding_alg = _decode_protobuf(
                message.content, saved_typedef, self.bbpb_config
            )
        elif flow_part.startswith("websocket"):
            # Websockets don't have a single typedef to edit
            # Instead, we are going to build a typedef based on all the messages
            if flow_part == "websocket-request":
                if not flow.websocket:
                    raise exceptions.CommandError(
                        f"Flow part is websocket-request, but flow is not a websocket"
                    )
                messages = [
                    message
                    for message in flow.websocket.messages
                    if message.from_client
                ]
            elif flow_part == "websocket-response":
                if not flow.websocket:
                    raise exceptions.CommandError(
                        f"Flow part is websocket-response, but flow is not a websocket"
                    )
                messages = [
                    message
                    for message in flow.websocket.messages
                    if not message.from_client
                ]
            if not messages:
                raise exceptions.CommandError(
                    f"Could not find any messages for flow part: {flow_part}"
                )
            message_hash = _message_hash(messages[0].content, messages[0], flow)
            saved_typedef = self.typedef_lookup.get(message_hash)
            try:
                typedef = saved_typedef
                message_jsons = []
                for message in messages:
                    message_json, typedef, encoding_alg = _decode_protobuf(
                        message.content, typedef, self.bbpb_config, fallback=False
                    )
                    message_jsons.append(message_json)
            except BlackboxProtobufException:
                typedef = {}
                message_jsons = []
                for message in messages:
                    message_json, typedef, encoding_alg = _decode_protobuf(
                        message.content, typedef, self.bbpb_config, fallback=False
                    )
                    message_jsons.append(message_json)
        else:
            raise exceptions.CommandError(f"Got unknown flow_part: {flow_part}")

        return typedef, message_hash

    @command.command("bbpb.del_type")
    @command.argument("typename", type=types.Choice("bbpb.options.known_types"))
    def bbpb_del_type(self, typename: str) -> None:
        if typename not in self.bbpb_config.known_types:
            raise exceptions.CommandError(f"Error: Type {typename} is not known")
        self.bbpb_config.known_types.pop(typename, None)
        keys_to_remove = [
            key for key, value in self.typedef_lookup.items() if value == typename
        ]
        for key in keys_to_remove:
            self.typedef_lookup.pop(key, None)
        self._save_project_file()
        self._refresh_view()

    @command.command("bbpb.options.edit_part")
    def bbpb_options_edit_part(self) -> Sequence[str]:
        flow = ctx.master.view.focus.flow
        if not flow:
            raise exceptions.CommandError("No flow selected.")

        # Prompts the user for the section to edit
        if flow.websocket:
            if len(flow.websocket.messages) > 0:
                return ["websocket"]
            else:
                return []
        elif flow.response:
            return [
                "request-body",
                "response-body",
            ]
        else:
            return ["request-body"]

    @command.command("bbpb.options.edit_type_part")
    def bbpb_options_edit_type_part(self) -> Sequence[str]:
        flow = ctx.master.view.focus.flow
        if flow.websocket:
            return ["websocket-request", "websocket-response"]
        if flow.response:
            return [
                "request-body",
                "response-body",
            ]
        return ["request-body"]

    @command.command("bbpb.options.known_types")
    def bbpb_options_known_types(self) -> Sequence[str]:
        typenames = list(self.bbpb_config.known_types.keys())
        return typenames + ["(clear)"]

    def _refresh_view(self):
        ctx.master.window.stacks[0].windows["flowview"].body.contentview_changed(None)

    @command.command("bbpb.project.load")
    def bbpb_project_load(self, project_file: str) -> None:
        # TODO would be good to have errors propagated here
        self._load_project_file(project_file)

    @command.command("bbpb.project.save")
    def bbpb_project_save(self, project_file: str) -> None:
        # TODO would be good to have errors propagated here
        self._save_project_file(project_file)


class BlackboxProtobufView(contentviews.View):
    name = "Blackbox Protobuf"

    def __init__(self, addon: BlackboxProtobufAddon):
        self.addon = addon

    def __call__(
        self,
        data: bytes,
        *,
        content_type: str | None = None,
        flow: flow.Flow | None = None,
        http_message: http.Message | None = None,
        **unknown_metadata,
    ) -> contentviews.TViewResult:
        # No support for TCP or UDP flows
        if not isinstance(flow, http.HTTPFlow):
            return None

        if len(data) == 0:
            return

        # message_hash is for looking up the appropiate typedef for this request, based on URL and type of message
        message_hash = _message_hash(data, http_message, flow)

        typedef = self.addon.typedef_lookup.get(message_hash)

        message, typedef_out, encoding_alg = _decode_protobuf(
            data, typedef, self.addon.bbpb_config
        )

        title = "Protobuf"
        if isinstance(typedef, str):
            title += f"  |  Type: {typedef}"
        else:
            title += f"  |  Type: anonymous"

        return title, contentviews.format_text(message)

    def render_priority(
        self,
        data: bytes,
        *,
        content_type: str | None = None,
        flow: flow.Flow | None = None,
        http_message: http.Message | None = None,
        **unknown_metadata,
    ) -> float:
        if content_type:
            if "protobuf" in content_type or "grpc" in content_type:
                return 2
            else:
                return 0
        # We don't know if we can decode protobuf or not, so let's elect
        # ourselves for all websockets
        if flow.websocket:
            return 1
        return 0


# This could be improved by taking is_request from the client for some cases
def _message_hash(
    data: bytes,
    message: http.Message | websocket.WebSocketMessage | None,
    flow: flow.Flow | None,
):
    if isinstance(message, http.Request):
        return f"request|{flow.request.url}"
    elif isinstance(message, http.Response):
        return f"response|{flow.request.url}"
    elif flow.websocket:
        if message is None or not isinstance(message, websocket.WebSocketMessage):
            # TODO this is really hacky and might be wasted cycles if there are
            # a lot of messages
            # Mitmproxy won't give us the WebSocketMessage message type for
            # content views, so we can't tell which direction it's going from
            # just the message.
            try:
                message = next(
                    (m for m in flow.websocket.messages if m.content == data)
                )
            except StopIteration:
                logging.warn(
                    "Message hashing couldn't find matching message in flow.websocket.messages"
                )
                message = None

        # Default to request if we never figured out the message
        if message and not message.from_client:
            return f"websocket-response|{flow.request.url}"
        else:
            return f"websocket-request|{flow.request.url}"
    else:
        logging.warn(
            f"BBPB content view got a view that was not websocket, request or response: {type(http_message)}"
        )
        return None


def _decode_protobuf(data, typedef, config, fallback=True):
    try:
        decoders = payloads.find_decoders(data)
        for decoder in decoders:
            try:
                protobuf_data, encoding_alg = decoder(data)
            except BlackboxProtobufException:
                continue

            try:
                message, typedef_out = blackboxprotobuf.lib.protobuf_to_json(
                    protobuf_data, typedef, config=config
                )

                return message, typedef_out, encoding_alg
            except BlackboxProtobufException as exc:
                if encoding_alg == "none":
                    raise exc
                continue
    except BlackboxProtobufException as exc:
        if typedef and fallback:
            return _decode_protobuf(data, {}, config)
        else:
            raise exc
    raise BlackboxProtobufException(
        'Failed to decode protobuf, but did not catch "none" decoder. This should never be hit'
    )


# This function spawns an editor for a text file. Once the file is saved, it
# calls `callback` with the text saved by the user. If `callback` throws an
# exception, this function will present the error to the user and allow the
# user to either keep editing the payload, reset the payload to the original
# text and edit again, or give up and exit with no changes
#
# The UI requires that all the logic on the modified text is embedded in a
# callback, I couldn't find any way to get a user choice without a callback
#
# This function is slightly cursed in that we have to use a custom chooser
# class to allow recursive chooser calls, otherwise it only allows a single
# prompt per command. Hopefully I can find a better way to handle this down the
# road.
T = TypeVar("T")


def _spawn_validating_editor(
    text: str, callback: Callable[[str], None], original_text: Optional[str] = None
) -> None:
    user_text = ctx.master.spawn_editor(text)
    if original_text is None:
        original_text = text

    try:
        callback(user_text)
        signals.pop_view_state.send()
    except Exception as exc:
        options = ["Continue Editing", "Reset Payload And Edit", "Exit"]

        def choose_callback(action: str):
            if action == "Continue Editing":
                # Keep editing the text that failed
                _spawn_validating_editor(user_text, callback, original_text)
            elif action == "Reset Payload And Edit":
                # Edit the original text instead
                _spawn_validating_editor(original_text, callback)
            elif action == "Exit":
                # Just return
                signals.pop_view_state.send()
                return
            else:
                raise Exception(
                    f"Got unknown option in validating editor menu: {action}"
                )

        signals.pop_view_state.send()
        ctx.master.overlay(
            RecursiveChooser(
                ctx.master,
                f"Error validating payload: {exc}",
                options,
                "",
                choose_callback,
            )
        )


# Below is the exact keypress implementation from `overlay.Chooser`, but
# without `signals.pop_view_state.send()` after the callbacks This allows us to
# spawn overlays (such as another chooser) from the callback function.
# With `overlay.Chooser`, it calls `signals.pop_view_state.send()` *after* the
# callback, which pops off the new chooser set by the callback
#
# This shifts responsibility for calling `signals.pop_view_state.send()` to the
# callback, which should call it on exit or before popping up a new chooser
class RecursiveChooser(overlay.Chooser):
    def keypress(self, size, key):
        key = self.master.keymap.handle_only("chooser", key)
        choice = self.walker.choice_by_shortcut(key)
        if choice:
            self.callback(choice)
            return
        if key == "m_select":
            self.callback(self.choices[self.walker.index])
            return
        elif key in ["q", "esc"]:
            signals.pop_view_state.send()
            return

        binding = self.master.keymap.get("global", key)
        # This is extremely awkward. We need a better way to match nav keys only.
        if binding and binding.command.startswith("console.nav"):
            self.master.keymap.handle("global", key)
        elif key in keymap.navkeys:
            return super().keypress(size, key)


addons = [BlackboxProtobufAddon()]



"""Module contains classes required to create Protobuf editor tabs."""

# Copyright (c) 2018-2023 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import six
import traceback
import base64
import zlib
import burp
import copy
import struct
import blackboxprotobuf
from javax.swing import JSplitPane, JPanel, JButton, BoxLayout, JOptionPane
from javax.swing import (
    Box,
    JTextField,
    JScrollPane,
    JList,
    ListSelectionModel,
    ListModel,
)
from javax.swing.event import ListSelectionListener, ListDataEvent, ListDataListener
from java.awt import Component, Dimension, FlowLayout
from java.awt.event import ActionListener
from javax.swing.border import EmptyBorder
from blackboxprotobuf.burp import user_funcs
from blackboxprotobuf.burp import typedef_editor
from blackboxprotobuf.lib import payloads
from blackboxprotobuf.lib.config import default as default_config
from blackboxprotobuf.lib.exceptions import (
    BlackboxProtobufException,
    DecoderException,
    EncoderException,
)

NAME_REGEX = re.compile(r"\A[a-zA-Z_][a-zA-Z0-9_]*\Z")


class ProtoBufEditorTabFactory(burp.IMessageEditorTabFactory):
    """Just returns instances of ProtoBufEditorTab"""

    def __init__(self, extender, callbacks):
        self._callbacks = callbacks
        self._extender = extender

    def createNewInstance(self, controller, editable):
        """Return new instance of editor tab for a new message"""
        return ProtoBufEditorTab(self._extender, controller, editable, self._callbacks)


class ProtoBufEditorTab(burp.IMessageEditorTab):
    """Tab in interceptor/repeater for editing protobuf message.

    Decodes the message to JSON and back for editing.
    The message type definition is attached to this object for re-encoding or editing.
    """

    def __init__(self, extension, controller, editable, callbacks):
        self._callbacks = callbacks
        self._extension = extension
        self._callbacks = extension.callbacks
        self._helpers = extension.helpers

        self._controller = controller

        self._text_editor = self._callbacks.createTextEditor()
        self._text_editor.setEditable(editable)
        self._editable = editable

        self._last_valid_type_index = None

        self._filtered_message_model = FilteredMessageModel(
            extension.known_message_model, self._callbacks
        )

        self._type_list_component = JList(self._filtered_message_model)
        self._type_list_component.setSelectionMode(ListSelectionModel.SINGLE_SELECTION)
        self._type_list_component.addListSelectionListener(TypeListListener(self))

        self._new_type_field = JTextField()

        self._component = JSplitPane(JSplitPane.HORIZONTAL_SPLIT)
        self._component.setLeftComponent(self._text_editor.getComponent())
        self._component.setRightComponent(self.createButtonPane())
        self._component.setResizeWeight(0.95)

        self._message_info = None
        self._payload_info = None
        self._last_good_msg = None  # (msg, typedef, source)
        self._decode_task = None

    def getTabCaption(self):
        """Return message tab caption"""
        return "Protobuf"

    def getMessage(self):
        """Transform the JSON format back to the binary protobuf message"""
        # Noticing this gets called twice for some reason
        # If we haven't finished decoding the message, cancel and return the original content
        if self._decode_task and (
            not self._decode_task.isDone() or self._decode_task.isCancelled()
        ):
            self._decode_task.cancel(True)
            self._decode_task = None
            # self._callbacks.printOutput(
            #    "Called getMessage before decode task was done, returning original content"
            # )
            return self._message_info.content()

        if self._last_good is None or not self.isModified():
            return self._message_info.content()

        success = False
        try:
            json_data = self._text_editor.getText().tostring()
            protobuf_data = blackboxprotobuf.protobuf_from_json(
                json_data, self._last_good.typedef
            )

            success = True
            self._payload_info.protobuf_data = protobuf_data
            return self._payload_info.generate_http(self._message_info, self._helpers)

        except Exception as exc:
            self._callbacks.printError(traceback.format_exc())

        if not success:
            try:
                protobuf_data = blackboxprotobuf.protobuf_from_json(
                    self._last_good.message, self._last_good.typedef
                )

                # Put the error here so that we only have one error to the user if the above encoding doesn't work
                JOptionPane.showMessageDialog(
                    self._component,
                    "Error encoding protobuf as-is. Reset data to previous good state: "
                    + str(exc),
                )

                success = True
                # Reset the message and protobuf data both
                self._text_editor.setText(self._last_good.message)
                self._payload_info.protobuf_data = protobuf_data
                return self._payload_info.generate_http(
                    self._message_info, self._helpers
                )
            except Exception as exc:
                self._callbacks.printError(traceback.format_exc())
                JOptionPane.showMessageDialog(
                    self._component,
                    "Error encoding protobuf. Setting data to the original message. Error: "
                    + str(exc),
                )
                self._text_editor.setText(self._message_info.content())
                return self._message_info.content()

    def _handle_protobuf(
        self, message_info, protobuf_data, message_type_in, typedef_source
    ):
        """
        Sets the protobuf message for the editor.
        """
        try:
            json_data, message_type = blackboxprotobuf.protobuf_to_json(
                protobuf_data, message_type_in
            )

            self._last_good = LastGoodData(json_data, message_type, typedef_source)
            self._filtered_message_model.set_new_data(protobuf_data)
            self._text_editor.setText(json_data)  # UI access
            success = True
        except Exception as exc:
            success = False
            self._callbacks.printError(
                "Got error decoding protobuf binary: " + traceback.format_exc()
            )

        # Bring out of exception handler to avoid nesting handlers
        if success:
            if typedef_source is not None:
                self.forceSelectType(typedef_source)
            else:
                self._type_list_component.clearSelection()
        elif len(message_type_in) == 0:
            self._callbacks.printError(
                "Error decoding protobuf with saved type, trying with empty type"
            )
            self._handle_protobuf(message_info, protobuf_data, {}, None)
        else:
            self._callbacks.printError("Error decoding protobuf with empty type")
            self._text_editor.setText("Error decoding protobuf")

    def setMessage(self, content, is_request):
        """
        Get the data from the request/response and parse into JSON.
        """
        # Run in a separate thread to avoid hanging the Burp UI
        # It has been observed that the Burp UI can hang when the message is large
        # and the decoding process takes a long time

        message_info = MessageInfo(content, is_request, self._helpers, self._controller)
        payload_info = PayloadInfo(message_info, self._helpers)
        message_type, typedef_source = self._get_saved_typedef(message_info)

        if (
            self._decode_task
            and not self._decode_task.isCancelled()
            and not self._decode_task.isDone()
        ):
            # If we're processing the same message as the running task
            # check message has
            # check protobuf data
            # TODO do we want to check typedef? Can switch away and back to restart it if we need to
            if (
                message_info.message_hash == self._message_info.message_hash
                and payload_info.raw_data == self._payload_info.raw_data
            ):
                # self._callbacks.printOutput(
                #    "Switched to tab that is still running with the same hash and payload. Not cancelling."
                # )
                return
            else:
                # self._callbacks.printOutput(
                #    "Have existing task that is still running, cancelling"
                # )
                # Cancel the old task
                self._decode_task.cancel(True)

        self._message_info = message_info
        self._payload_info = payload_info
        self._last_good_msg = None
        self._decode_task = None

        self._text_editor.setText("Please wait...")

        def run():
            try:
                decoders = payloads.find_decoders(payload_info.raw_data)
                for decoder in decoders:
                    try:
                        protobuf_data, encoding_alg = decoder(payload_info.raw_data)
                    except BlackboxProtobufException:
                        continue

                    try:
                        self._handle_protobuf(
                            message_info,
                            protobuf_data,
                            message_type,
                            typedef_source,
                        )

                        # Payload successfully decoded, so we probably have the right wrapper for the payload
                        self._payload_info.protobuf_data = protobuf_data
                        self._payload_info.encoding_alg = encoding_alg

                        return
                    except BlackboxProtobufException:
                        if encoding_alg == "none":
                            # Reraise the exception to the parent context and halt decoding
                            six.reraise(*sys.exc_info())
                        continue

            except Exception as ex:
                # Catch all, otherwise it disappears
                self._text_editor.setText("Error decoding protobuf")
                self._callbacks.printError("Error decoding protobuf: %s" % ex)

        self._decode_task = self._extension.thread_executor.submit(run)

    def getSelectedData(self):
        """Get text currently selected in message"""
        return self._text_editor.getSelectedText()

    def getUiComponent(self):
        """Return Java AWT component for this tab"""
        return self._component

    def isEnabled(self, content, is_request):
        """Try to detect a protobuf in the message to enable the tab.

        Defaults to content-type header of 'x-protobuf'. User overridable in
        `user_funcs.py`
        """
        # TODO implement some more default checks
        if is_request:
            info = self._helpers.analyzeRequest(content)
        else:
            info = self._helpers.analyzeResponse(content)

        if "detect_protobuf" in dir(user_funcs):
            result = user_funcs.detect_protobuf(
                content, is_request, info, self._helpers
            )
            if result is not None:
                return result

        # Bail early if there is no body
        if info.getBodyOffset() == len(content):
            return False

        protobuf_content_types = [
            "protobuf",
            "grpc",
        ]
        # Check all headers for x-protobuf
        for header in info.getHeaders():
            if "content-type" in header.lower():
                for protobuf_content_type in protobuf_content_types:
                    if protobuf_content_type in header.lower():
                        return True

        return False

    def isModified(self):
        """Return if the message was modified"""
        return self._text_editor.isTextModified()

    def createButtonPane(self):
        """Create a new button pane for the message editor tab"""
        self._button_listener = EditorButtonListener(self)

        panel = JPanel()
        panel.setLayout(BoxLayout(panel, BoxLayout.Y_AXIS))
        panel.setBorder(EmptyBorder(5, 5, 5, 5))

        panel.add(Box.createRigidArea(Dimension(0, 5)))
        type_scroll_pane = JScrollPane(self._type_list_component)
        type_scroll_pane.setMaximumSize(Dimension(200, 100))
        type_scroll_pane.setMinimumSize(Dimension(150, 100))
        panel.add(type_scroll_pane)
        panel.add(Box.createRigidArea(Dimension(0, 3)))

        new_type_panel = JPanel()
        new_type_panel.setLayout(BoxLayout(new_type_panel, BoxLayout.X_AXIS))
        new_type_panel.add(self._new_type_field)
        new_type_panel.add(Box.createRigidArea(Dimension(3, 0)))
        new_type_panel.add(
            self.createButton(
                "New", "new-type", "Save this message's type under a new name"
            )
        )
        new_type_panel.setMaximumSize(Dimension(200, 20))
        new_type_panel.setMinimumSize(Dimension(150, 20))

        panel.add(new_type_panel)

        button_panel = JPanel()
        button_panel.setLayout(FlowLayout())
        if self._editable:
            button_panel.add(
                self.createButton(
                    "Validate", "validate", "Validate the message can be encoded."
                )
            )
        button_panel.add(
            self.createButton("Edit Type", "edit-type", "Edit the message type")
        )
        button_panel.add(
            self.createButton(
                "Reset Message", "reset", "Reset the message and undo changes"
            )
        )
        button_panel.add(
            self.createButton(
                "Clear Type", "clear-type", "Reparse the message with an empty type"
            )
        )
        button_panel.setMinimumSize(Dimension(100, 200))
        button_panel.setPreferredSize(Dimension(200, 1000))

        panel.add(button_panel)

        return panel

    def createButton(self, text, command, tooltip):
        """Create a new button with the given text and command"""
        button = JButton(text)
        button.setAlignmentX(Component.CENTER_ALIGNMENT)
        button.setActionCommand(command)
        button.addActionListener(self._button_listener)
        button.setToolTipText(tooltip)
        return button

    def validateMessage(self):
        """Callback for validate button. Attempts to encode the message with
        the current type definition
        """
        try:
            json_data = self._text_editor.getText().tostring()
            typedef = self._last_good.typedef
            protobuf_data = blackboxprotobuf.protobuf_from_json(json_data, typedef)
            # If it works, save the message
            # Don't need to save typeddef because we're using the one from lastgood
            self._last_good.message = json_data
            self._payload_info.protobuf_data = protobuf_data

        except Exception as exc:
            JOptionPane.showMessageDialog(
                self._component,
                "Got exception while trying to encode the message: " + str(exc),
            )
            self._callbacks.printError(traceback.format_exc())

    def resetMessage(self):
        """Drop any changes and revert to the last good message. Callback for
        "reset" button
        """

        self._text_editor.setText(self._last_good.message)

    def forceSelectType(self, typename):
        index = self._filtered_message_model.get_type_index(typename)
        if index is not None:
            self._last_valid_type_index = index
            self._type_list_component.setSelectedIndex(index)

    def updateTypeSelection(self):
        """Apply a new typedef based on the selected type in the type list"""
        # TODO It sucks that we lose the anonymous type if we accidentally
        # click a type. Maybe we should have an entry in the cached type in the
        # list?
        # Or have a warning before switching
        # Or require a click + a button press?

        # Check if something is selected
        if self._type_list_component.isSelectionEmpty():
            self._last_valid_type_index = None
            self._extension.saved_types.pop(self._message_info.message_hash, None)
            return

        # TODO won't actually work right if we delete the type we're using a
        # new type is now in the index
        if self._last_valid_type_index == self._type_list_component.getSelectedIndex():
            # hasn't actually changed since last time we tried
            # otherwise can trigger a second time when we call setSelectedIndex below on failure
            return

        type_name = self._type_list_component.getSelectedValue()
        # try to catch none here...
        if not type_name or type_name not in default_config.known_types:
            return

        try:
            self.applyType(default_config.known_types[type_name], type_name)
        except BlackboxProtobufException as exc:
            self._callbacks.printError(traceback.format_exc())

            if isinstance(exc, EncoderException):
                JOptionPane.showMessageDialog(
                    self._component,
                    "Error encoding protobuf with previous type: %s" % (exc),
                )
            elif isinstance(exc, DecoderException):
                JOptionPane.showMessageDialog(
                    self._component,
                    "Error encoding protobuf with type %s: %s" % (type_name, exc),
                )
                # decoder exception means it doesn't match the message that was sucessfully encoded by the prev type
                self._filtered_message_model.remove_type(type_name)

            if self._last_valid_type_index is not None:
                type_name = self._type_list_component.setSelectedIndex(
                    self._last_valid_type_index
                )
            else:
                self._type_list_component.clearSelection()
            return

        self._extension.saved_types[self._message_info.message_hash] = type_name
        self._last_valid_type_index = self._type_list_component.getSelectedIndex()

    def editType(self, typedef, source):
        """Apply and save the new typedef"""
        # Try to apply the type first
        try:
            # TODO Background this like with handle_protobuf? I think we need
            # to be more confident that we can validate the typedef without
            # decoding first
            # The decoding here throws an exception that will prevent us from
            # closing the typedef editor window if it's not valid
            self.applyType(typedef, source)
        except BlackboxProtobufException as exc:
            self._callbacks.printError("Got exception trying to apply edited typedef.")
            JOptionPane.showMessageDialog(
                self._component,
                "Error decoding the protobuf with the new type: %s" % (exc),
            )
            return

        if source is None:
            # Anonymous typedef tied to message hash
            # save the typedef
            self._extension.saved_types[self._message_info.message_hash] = typedef
        else:
            # Named typedef
            # save under known typedefs and save the name in source
            default_config.known_types[source] = typedef
            self._extension.saved_types[self._message_info.message_hash] = source

    def applyType(self, typedef, source):
        """Apply a new typedef to the message. Throws an exception if type is invalid."""
        # Convert to protobuf as old type and re-interpret as new type
        old_typedef = self._last_good.typedef
        json_data = self._text_editor.getText().tostring()

        protobuf_data = blackboxprotobuf.protobuf_from_json(json_data, old_typedef)
        self._payload_info.protobuf_data = protobuf_data

        new_json, new_typedef = blackboxprotobuf.protobuf_to_json(
            protobuf_data, typedef
        )

        self._last_good = LastGoodData(new_json, new_typedef, source)

        self._filtered_message_model.set_new_data(protobuf_data)
        self._text_editor.setText(str(new_json))
        # We do not try to remember the type here, this should be handled by the caller

    def saveAsNewType(self):
        """Copy the current type into known_messages"""

        name = self._new_type_field.getText().strip()
        if not NAME_REGEX.match(name):
            JOptionPane.showMessageDialog(
                self._component,
                "%s is not a valid "
                "message name. Message names should be alphanumeric." % name,
            )
            return
        if name in default_config.known_types:
            JOptionPane.showMessageDialog(
                self._component, "Message name %s is " "already taken." % name
            )
            return

        typedef = self._last_good.typedef

        # Do a deep copy on the dictionary so we don't accidentally modify others
        default_config.known_types[name] = copy.deepcopy(typedef)
        self._last_good.source = name  # remember the source, typedef is still the same

        # update the list of messages. This should trickle down to known message model
        self._extension.known_message_model.addElement(name)
        self._new_type_field.setText("")
        self._extension.saved_types[self._message_info.message_hash] = name

        # force select our new type
        self.forceSelectType(name)

    def clearType(self):
        self.applyType({}, None)
        self._type_list_component.clearSelection()
        self._new_type_field.setText("")
        self._extension.saved_types.pop(self._message_info.message_hash, None)

    def open_typedef_window(self):
        typedef = self._last_good.typedef
        source = self._last_good.source
        self._extension.open_typedef_editor(typedef, source, self.editType)

    def _get_saved_typedef(self, message_info):
        # Get the typedef we have saved for this message
        # It can be anonymous, but saved based on the message hash
        # Or it can be named
        # Return typedef, typename tuple
        if message_info.message_hash in self._extension.saved_types:
            saved_type = self._extension.saved_types[message_info.message_hash]
            if isinstance(saved_type, dict):
                return saved_type, None
            elif saved_type in default_config.known_types:
                typename = saved_type
                typedef = default_config.known_types[typename]
                return typedef, typename
            else:
                # We had a type, but it isn't a dict and not in known types
                # Error, so clear
                self._extension.saved_types.pop(message_info.message_hash, None)
                self._callbacks.printError(
                    "Found unknown saved type: %s for %s"
                    % (saved_type, message_info.message_hash)
                )
        else:
            return {}, None


class EditorButtonListener(ActionListener):
    """Callback listener for buttons in the message editor tab"""

    def __init__(self, editor_tab):
        self._editor_tab = editor_tab

    def actionPerformed(self, event):
        """Called when when a button in the message editor is pressed"""
        if event.getActionCommand() == "validate":
            self._editor_tab.validateMessage()
        elif event.getActionCommand() == "reset":
            self._editor_tab.resetMessage()
        elif event.getActionCommand() == "edit-type":
            self._editor_tab.open_typedef_window()
        elif event.getActionCommand() == "new-type":
            self._editor_tab.saveAsNewType()
        elif event.getActionCommand() == "clear-type":
            self._editor_tab.clearType()


class TypeListListener(ListSelectionListener):
    """Callback listener for when a new type is selected form the list"""

    def __init__(self, editor_tab):
        self._editor_tab = editor_tab

    def valueChanged(self, event):
        if event.getValueIsAdjusting():
            return
        self._editor_tab.updateTypeSelection()


class FilteredMessageModel(ListModel, ListDataListener):
    """Listens to a java ListModel and keeps a subset with just valid types
    for a message.
    """

    def __init__(self, parent, callbacks):
        self._callbacks = callbacks
        self._data = None
        self._parent_model = parent
        self._listeners = []

        # list of types in the parent model
        self._types = []
        self._parent_types = set(parent.elements())
        self._rejected_types = set()
        self._working_types = set()

        self._parent_model.addListDataListener(self)

    def set_new_data(self, data):
        self._data = data

        # clear caches
        self._working_types.clear()
        self._rejected_types.clear()

        # recheck all the types with the new data
        for typename in self._types[:]:
            if not self._check_type(typename):
                removed_index = self._types.index(typename)
                self._types.remove(typename)
                event = ListDataEvent(
                    self, ListDataEvent.INTERVAL_REMOVED, removed_index, removed_index
                )
                self._send_event(event)

        interval_start = len(self._types)
        for typename in self._parent_types:
            if typename not in self._types and self._check_type(typename):
                self._types.append(typename)

        if len(self._types) > interval_start:
            event = ListDataEvent(
                self, ListDataEvent.INTERVAL_ADDED, interval_start, len(self._types) - 1
            )
            self._send_event(event)

    def get_type_index(self, typename):
        if typename in self._types:
            return self._types.index(typename)
        return None

    def remove_type(self, typename):
        # if we failed to apply a type, then this lets us remove it
        if typename not in self._types:
            return

        type_index = self._types.index(typename)
        # delete by index because we have it
        del self._types[type_index]
        self._working_types.remove(typename)
        self._rejected_types.add(typename)

        event = ListDataEvent(
            self, ListDataEvent.INTERVAL_REMOVED, type_index, type_index
        )
        self._send_event(event)

    def update_types(self):
        new_parent_types = set(self._parent_model.elements())
        added_parent_types = new_parent_types - self._parent_types
        removed_parent_types = self._parent_types - new_parent_types
        self._parent_types = new_parent_types

        for type_name in removed_parent_types:
            if type_name in self._types:
                removed_index = self._types.index(type_name)
                self._types.remove(type_name)
                event = ListDataEvent(
                    self, ListDataEvent.INTERVAL_REMOVED, removed_index, removed_index
                )
                self._send_event(event)

        interval_start = len(self._types)
        for type_name in added_parent_types:
            if type_name not in self._types and self._check_type(type_name):
                self._types.append(type_name)

        # not sure how much effort we want to put into the events. could just mark everything as changed all the time
        if len(self._types) > interval_start:
            # if we didn't remove anything, then  just issue an added event?
            event = ListDataEvent(
                self, ListDataEvent.INTERVAL_ADDED, interval_start, len(self._types) - 1
            )
            self._send_event(event)

    def _send_event(self, event):
        event_type = event.getType()
        if event_type == ListDataEvent.CONTENTS_CHANGED:
            for listener in self._listeners:
                listener.contentsChanged(event)
        elif event_type == ListDataEvent.INTERVAL_ADDED:
            for listener in self._listeners:
                listener.intervalAdded(event)
        elif event_type == ListDataEvent.INTERVAL_REMOVED:
            for listener in self._listeners:
                listener.intervalRemoved(event)

    def _check_type(self, typename):
        # TODO this hangs the UI as well
        # TODO would be better to check by comparing typedefs instead of trying
        # to decode
        if typename in self._rejected_types:
            return False
        if typename in self._working_types:
            return True

        # if we don't have data yet, just quit early
        if not self._data:
            return False
        if typename not in default_config.known_types:
            return False
        typedef = default_config.known_types[typename]
        try:
            _, _ = blackboxprotobuf.protobuf_to_json(self._data, typedef)
        except BlackboxProtobufException as exc:
            self._callbacks.printError(traceback.format_exc())
            self._rejected_types.add(typename)
            return False
        self._working_types.add(typename)
        return True

    def addListDataListener(self, listener):
        self._listeners.append(listener)

    def getElementAt(self, i):
        return self._types[i]

    def getSize(self):
        return len(self._types)

    def removeListDataListener(self, listener):
        self._listeners.remove(listener)

    # data listener stuff
    def contentsChanged(self, event):
        self.update_types()

    def intervalAdded(self, event):
        self.update_types()

    def intervalRemoved(self, event):
        self.update_types()


class MessageInfo:
    """This class parses the data we get from burp for us to use throughout the processing"""

    def __init__(self, content, is_request, helpers, controller):
        self.is_request = is_request

        if is_request:
            self.request = content
        else:
            self.request = controller.getRequest()

            self.response = content
            self.response_content_info = helpers.analyzeResponse(content)

        self.request_content_info = helpers.analyzeRequest(
            controller.getHttpService(), self.request
        )

        self.message_hash = self._message_hash(helpers)

    def content(self):
        if self.is_request:
            return self.request
        else:
            return self.response

    def content_info(self):
        if self.is_request:
            return self.request_content_info
        else:
            return self.response_content_info

    def _message_hash(self, helpers):
        """Compute an "identifier" for the message which is used for sticky
        type definitions. User modifiable
        """
        message_hash = None
        if "hash_message" in dir(user_funcs):
            message_hash = user_funcs.hash_message(
                self.content(),
                self.is_request,
                self.content_info(),
                helpers,
                self.request,
                self.request_content_info,
            )
        if message_hash is None:
            # Base it off just the URL and request/response
            url = self.request_content_info.getUrl()
            message_hash = ":".join(
                [url.getAuthority(), url.getPath(), str(self.is_request)]
            )

        return message_hash


class PayloadInfo:
    """This class stores the latest payload data and functions to transform it to or from a HTTP message"""

    def __init__(self, message_info, helpers):
        self.raw_data = None  # Raw data from payload
        # These attributes must be set from outside this payload
        self.encoding_alg = None
        self.protobuf_data = None  # last known good encoded protobuf payload
        self.parse_http(message_info, helpers)

    def parse_http(self, message_info, helpers):
        raw_data = None
        if "get_protobuf_data" in dir(user_funcs):
            raw_data = user_funcs.get_protobuf_data(
                message_info.content(),
                message_info.is_request,
                message_info.content_info(),
                helpers,
                message_info.request,
                message_info.request_content_info,
            )
        if raw_data is None:
            raw_data = message_info.content()[
                message_info.content_info().getBodyOffset() :
            ].tostring()
        self.raw_data = raw_data

    def generate_http(self, message_info, helpers):
        if "set_protobuf_data" in dir(user_funcs):
            result = user_funcs.set_protobuf_data(
                self.protobuf_data,
                message_info.content(),
                message_info.is_request,
                message_info.content_info(),
                helpers,
                message_info.request,
                message_info.request_content_info,
            )
            if result is not None:
                return result

        if self.protobuf_data is None:
            raise BlackboxProtobufException(
                "Error generating HTTP body. PayloadInfo does not have valid protobuf data to encode"
            )
        raw_data = payloads.encode_payload(self.protobuf_data, self.encoding_alg)
        headers = message_info.content_info().getHeaders()
        return helpers.buildHttpMessage(headers, str(raw_data))


class LastGoodData:
    """This class stores data about the last now valid combination of message and typedef"""

    def __init__(self, message, typedef, source):
        self.message = message
        self.typedef = typedef
        self.source = source



"""These functions provide hooks into the blackboxprotobuf Burp extension,
allowing certain functionality to be customized in order to handle non-standard
applications.
"""

# Copyright (c) 2018-2023 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Documentation for Common parameters:
#   content -- Binary content of the request
#   is_request -- boolean, True for a request, False for a response
#   content_info -- RequestInfo or ResponseInfo object, See
#       https://portswigger.net/Burp/extender/api/burp/IRequestInfo.html and
#       https://portswigger.net/Burp/extender/api/burp/IResponseInfo.html
#   helpers -- Burp extension helpers,
#       https://portswigger.net/Burp/extender/api/burp/IExtensionHelpers.html
#   request/request_content_info -- If called on a response, send the
#       corresponding request (useful for retrieving URL parameters)

# Useful functionality:
#    URL parameters:
#        for param in content_info.getParmeters():
#            if param.getName() == 'type':
#                ...
#    Headers:
#        if 'content-type' in content_info.getHeaders():
#            ...
#    Request Body:
#       body = content[content_info.getBodyOffset():].tostring()
#    Setting paramater:
#       import burp.IParameter
#       body = helpers.updateParameter(
#                   content,
#                   helpers.buildParameter('message',
#                                          protobuf_data,
#                                          IParameter.PARAM_URL))


def detect_protobuf(content, is_request, content_info, helpers):
    """Customize protobuf detection with a request or a response. Should return True if it is a protobuf,
    False if it definitely not a protobuf and should not try to decode, or None
    to fallback to the standard content-type header detection.

    This can be used to add protobuf detection based on different headers or
    customer application functionality. You can also use "return True" to try
    to decode every request/response.
    """
    pass


def get_protobuf_data(
    content, is_request, content_info, helpers, request=None, request_content_info=None
):
    """Customize how the protobuf data is retrieved from the request. By
    default, it is assumed the body of the request/response contains a
    protobuf payload and tries to detect if it's compressed.

    This function can be used if the payload is in a non-standard location or
    has a non-standard encoding. It should return the raw protobuf bytes from
    the message.
    """
    pass


def set_protobuf_data(
    protobuf_data,
    content,
    is_request,
    content_info,
    helpers,
    request=None,
    request_content_info=None,
):
    """Customize how the protobuf data is set in request/response. This
    function is used to configure where in the request the protobuf data is
    placed and how it is encoded if the message is modified. For example, this
    could set a query parameter with base64 encoded data if that is what the
    application expects. This function is the mirror to `get_protobuf_data` and
    if one function is modified, then the other should as well.

    This is expected to return bytes representing the HTTP headers and body,
    such as that returned by the `buildHttpMessage` function of the Burp
    helpers
    (https://portswigger.net/burp/extender/api/burp/iextensionhelpers.html#buildHttpMessage)
    """
    pass


def hash_message(
    content, is_request, content_info, helpers, request=None, request_content_info=None
):
    """Blackbox protobuf remembers which type definition to use for each
    request/response by saving them in a dictionary, keyed by a 'message hash'.
    If a message has the same hash, it will try to re-use the same type
    definition. By default, this is the request path and true or false
    depending whether it is a request or response.

    If there isn't a one-to-one mapping of message types to URLs, then this
    function can be used to customize what parts of the request to use for
    identification (eg. a message type header or parameter).

    This function should return a string.
    """
    pass



# Copyright (c) 2018-2023 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
import inspect

# Add correct directory to sys.path
_BASE_DIR = os.path.abspath(
    os.path.dirname(inspect.getfile(inspect.currentframe())) + "../../../"
)

sys.path.insert(0, _BASE_DIR + "/lib/")
sys.path.insert(0, _BASE_DIR + "/burp/deps/six/")
sys.path.insert(0, _BASE_DIR + "/burp/deps/protobuf/python/")

# extend_path looks for other 'blackboxprotobuf' modules in the sys.path and
# adds them to __path__
from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)


# Hack to fix loading protobuf libraries within Jython. See https://github.com/protocolbuffers/protobuf/issues/7776
def fix_protobuf():
    import six

    u = six.u

    def new_u(s):
        if s == r"[\ud800-\udfff]":
            # Don't match anything
            return "$^"
        else:
            return u(s)

    six.u = new_u


fix_protobuf()

# mirror what we do in lib so we can use blackboxprotobuf.<function>
from blackboxprotobuf.lib.api import *



""" TypeDefinitionTab is a top-level Burp Suite tab which allows saved/named
    types to be added/modified at any time. It also adds options for
    importing/exporting protobuf types to .json files.
"""

# Copyright (c) 2018-2023 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import re
import json
import burp
import traceback
import blackboxprotobuf
from javax.swing import JSplitPane, JScrollPane, JPanel, JButton, BoxLayout, Box
from javax.swing import JOptionPane, JList, ListSelectionModel, JFileChooser
from javax.swing.filechooser import FileNameExtensionFilter
from java.awt import Component, Dimension
from java.awt.event import ActionListener
from javax.swing.border import EmptyBorder
from blackboxprotobuf.lib.api import sort_typedef
from blackboxprotobuf.lib.config import default as default_config

from blackboxprotobuf.burp import typedef_editor

# TODO put these in one place
NAME_REGEX = re.compile(r"\A[a-zA-Z_][a-zA-Z0-9_]*\Z")


class TypeDefinitionTab(burp.ITab):
    """Implements an interface for editing known message type definitions."""

    def __init__(self, extension, burp_callbacks):
        self._burp_callbacks = burp_callbacks
        self._extension = extension

        self._type_list_component = JList(extension.known_message_model)
        self._type_list_component.setSelectionMode(
            ListSelectionModel.MULTIPLE_INTERVAL_SELECTION
        )

        self._component = JPanel()
        self._component.setLayout(BoxLayout(self._component, BoxLayout.Y_AXIS))

        splitPane = JSplitPane(JSplitPane.HORIZONTAL_SPLIT)
        splitPane.setRightComponent(JScrollPane(self._type_list_component))
        splitPane.setLeftComponent(self.createButtonPane())
        splitPane.setResizeWeight(0.03)
        splitPane.setMaximumSize(Dimension(1000, 1000))

        self._component.add(splitPane)
        self._component.add(Box.createVerticalGlue())
        self._component.setBorder(EmptyBorder(10, 10, 10, 10))

    def getTabCaption(self):
        """Returns name on tab"""
        return "Protobuf Type Editor"

    def getUiComponent(self):
        """Returns Java AWT component for tab"""
        return self._component

    def createButtonPane(self):
        """Create AWT window panel for buttons"""
        self._button_listener = TypeDefinitionButtonListener(self)

        panel = JPanel()
        panel.setLayout(BoxLayout(panel, BoxLayout.Y_AXIS))

        panel.add(Box.createRigidArea(Dimension(0, 5)))
        panel.add(self.createButton("Add", "new-type", "Create a new type"))
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(self.createButton("Edit", "edit-type", "Edit the selected type"))
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(
            self.createButton("Rename", "rename-type", "Rename the selected type")
        )
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(
            self.createButton("Remove", "delete-type", "Delete all selected types")
        )
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(
            self.createButton(
                "Save All Types To File",
                "save-types",
                "Save all known types as JSON to a file",
            )
        )
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(
            self.createButton(
                "Load All Types From File", "load-types", "Load types from JSON file"
            )
        )
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(
            self.createButton(
                "Export All types As .proto",
                "export-proto",
                "Export all types as .proto",
            )
        )
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(
            self.createButton(
                "Import .proto", "import-proto", "Import types from a .proto file"
            )
        )
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        return panel

    def createButton(self, text, command, tooltip):
        """Generate new button with the given text and command string"""
        button = JButton(text)
        button.setAlignmentX(Component.CENTER_ALIGNMENT)
        button.setActionCommand(command)
        button.addActionListener(self._button_listener)
        button.setToolTipText(tooltip)
        return button

    def save_callback(self, typedef, name):
        """Save typedef and update list for a given message name"""
        if name not in default_config.known_types:
            self._extension.known_message_model.addElement(name)

        default_config.known_types[name] = typedef

    def add_typedef(self):
        type_name = JOptionPane.showInputDialog("Enter new name")

        # Error out if already defined
        if type_name in default_config.known_types:
            JOptionPane.showMessageDialog(
                self._component,
                'Message type "%s" already exists' % type_name,
            )
            return

        self._extension.open_typedef_editor({}, type_name, self.save_callback)

    def edit_typedef(self):
        list_component = self._type_list_component
        if list_component.isSelectionEmpty():
            return

        type_name = list_component.getSelectedValue()
        message_type = default_config.known_types[type_name]

        self._extension.open_typedef_editor(
            sort_typedef(message_type), type_name, self.save_callback
        )


class TypeDefinitionButtonListener(ActionListener):
    """Callback listener for buttons in the TypeDefinition interface"""

    def __init__(self, typedef_tab):
        self._typedef_tab = typedef_tab

    def actionPerformed(self, event):
        """Called when a button is pressed."""
        if event.getActionCommand() == "new-type":
            self._typedef_tab.add_typedef()

        elif event.getActionCommand() == "edit-type":
            self._typedef_tab.edit_typedef()

        elif event.getActionCommand() == "rename-type":
            list_component = self._typedef_tab._type_list_component
            # Check if something is selected
            if list_component.isSelectionEmpty():
                return

            # Get's only the first value
            previous_type_name = list_component.getSelectedValue()
            new_type_name = JOptionPane.showInputDialog(
                "Enter new name for %s:" % previous_type_name
            )
            if new_type_name in default_config.known_types:
                JOptionPane.showMessageDialog(
                    self._typedef_tab._component,
                    'Message type "%s" already exists' % new_type_name,
                )
                return
            if previous_type_name not in default_config.known_types:
                JOptionPane.showMessageDialog(
                    self._typedef_tab._component,
                    'Message type "%s" does not exist' % previous_type_name,
                )
                return
            if not NAME_REGEX.match(new_type_name):
                JOptionPane.showMessageDialog(
                    self._typedef_tab._component,
                    'Message type name "%s" is not valid.' % new_type_name,
                )
                return
            typedef = default_config.known_types[previous_type_name]
            default_config.known_types[new_type_name] = typedef
            del default_config.known_types[previous_type_name]
            # TODO should manage this centrally somewhere
            self._typedef_tab._extension.refresh_message_model()
            for key, typename in self._typedef_tab._extension.saved_types.items():
                if typename == previous_type_name:
                    self._typedef_tab._extension.saved_types[key] = new_type_name

        elif event.getActionCommand() == "delete-type":
            list_component = self._typedef_tab._type_list_component
            # Check if something is selected
            if list_component.isSelectionEmpty():
                return

            type_names = list_component.getSelectedValuesList()
            # TODO Confirm delete?
            for type_name in type_names:
                del default_config.known_types[type_name]
            self._typedef_tab._extension.refresh_message_model()

        elif event.getActionCommand() == "save-types":
            chooser = JFileChooser()
            chooser.setFileFilter(
                FileNameExtensionFilter("JSON Type Definition", ["json"])
            )
            chooser.setMultiSelectionEnabled(False)

            action = chooser.showSaveDialog(self._typedef_tab.getUiComponent())
            if (
                action == JFileChooser.CANCEL_OPTION
                or action == JFileChooser.ERROR_OPTION
            ):
                return

            file_name = chooser.getSelectedFile().getCanonicalPath()
            ext = os.path.splitext(file_name)[1]
            if ext != ".json":
                # Add json extension if it doesn't have one
                file_name += ".json"

            with open(file_name, "w+") as selected_file:
                json.dump(
                    default_config.known_types,
                    selected_file,
                    indent=4,
                    sort_keys=True,
                )

        elif event.getActionCommand() == "load-types":
            chooser = JFileChooser()
            chooser.setFileFilter(
                FileNameExtensionFilter("JSON Type Definition", ["json"])
            )
            chooser.setMultiSelectionEnabled(False)

            action = chooser.showOpenDialog(self._typedef_tab.getUiComponent())
            if (
                action == JFileChooser.CANCEL_OPTION
                or action == JFileChooser.ERROR_OPTION
            ):
                return

            file_name = chooser.getSelectedFile().getCanonicalPath()
            types = {}
            with open(file_name, "r") as selected_file:
                types = json.load(selected_file)
            for key, value in types.items():
                # check to make sure we don't nuke existing messages
                if key in default_config.known_types:
                    overwrite = (
                        JOptionPane.showConfirmDialog(
                            self._typedef_tab._component,
                            "Message %s already saved. Overwrite?" % key,
                        )
                        == 0
                    )
                    if not overwrite:
                        continue
                default_config.known_types[key] = value
            self._typedef_tab._extension.refresh_message_model()
        elif event.getActionCommand() == "export-proto":
            chooser = JFileChooser()
            chooser.setFileFilter(
                FileNameExtensionFilter("Protobuf Type Definition", ["proto"])
            )
            chooser.setMultiSelectionEnabled(False)

            action = chooser.showSaveDialog(self._typedef_tab.getUiComponent())
            if (
                action == JFileChooser.CANCEL_OPTION
                or action == JFileChooser.ERROR_OPTION
            ):
                return

            file_name = chooser.getSelectedFile().getCanonicalPath()
            ext = os.path.splitext(file_name)[1]
            if ext == "":
                # No extension, add .proto
                file_name += ".proto"

            if os.path.exists(file_name):
                # 0 is the YES option
                overwrite = (
                    JOptionPane.showConfirmDialog(
                        self._typedef_tab._component,
                        "File %s already exists. Overwrite?" % file_name,
                    )
                    == 0
                )
                if not overwrite:
                    return
                print("overwriting file: %s" % file_name)
            try:
                blackboxprotobuf.export_protofile(default_config.known_types, file_name)
            except Exception as exc:
                self._typedef_tab._burp_callbacks.printError(traceback.format_exc())
                JOptionPane.showMessageDialog(
                    self._typedef_tab._component,
                    "Error saving .proto file: " + str(exc),
                )

        elif event.getActionCommand() == "import-proto":
            chooser = JFileChooser()
            chooser.setFileFilter(
                FileNameExtensionFilter("Protobuf Type Definition", ["proto"])
            )
            chooser.setMultiSelectionEnabled(False)

            action = chooser.showOpenDialog(self._typedef_tab.getUiComponent())
            if (
                action == JFileChooser.CANCEL_OPTION
                or action == JFileChooser.ERROR_OPTION
            ):
                return

            file_name = chooser.getSelectedFile().getCanonicalPath()
            if not os.path.exists(file_name):
                self._typedef_tab._burp_callbacks.printError(
                    "Attempted to import %s, but the file does not exist." % file_name
                )
                JOptionPane.showMessageDialog(
                    self._typedef_tab._component,
                    "File %s does not exist to import." + str(exc),
                )
                return
            try:
                new_typedefs = blackboxprotobuf.import_protofile(
                    file_name, save_to_known=False
                )
                for key, value in new_typedefs.items():
                    # check to make sure we don't nuke existing messages
                    if key in default_config.known_types:
                        overwrite = (
                            JOptionPane.showConfirmDialog(
                                self._typedef_tab._component,
                                "Message %s already saved. Overwrite?" % key,
                            )
                            == 0
                        )
                        if not overwrite:
                            continue
                    else:
                        self._typedef_tab._extension.known_message_model.addElement(key)
                    default_config.known_types[key] = value
            except Exception as exc:
                self._typedef_tab._burp_callbacks.printError(traceback.format_exc())
                JOptionPane.showMessageDialog(
                    self._typedef_tab._component,
                    "Error loading .proto file: " + str(exc),
                )



"""Contains classes for a window for editing type definitions. Called from both
the message editor tab and from the suite tabs.
"""

# Copyright (c) 2018-2023 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import traceback
import json
import blackboxprotobuf
from java.awt import Component, Dimension, Frame
from java.awt.event import ActionListener, WindowEvent
from javax.swing import (
    JSplitPane,
    JPanel,
    JButton,
    BoxLayout,
    JOptionPane,
    JDialog,
    Box,
    JTextField,
)


class TypeEditorWindow(JDialog):
    """New free-standing window for editing a specified type definition. Will
    callback into the calling class when the type is saved
    """

    def __init__(self, burp_callbacks, typedef, source, callback):
        burp_window = None
        for frame in Frame.getFrames():
            if "Burp Suite" in frame.getName():
                burp_window = frame
                break
        JDialog.__init__(self, burp_window)
        self._burp_callbacks = burp_callbacks
        self._type_callback = callback
        self.setSize(1000, 700)

        self._original_typedef = typedef
        self._type_source = source
        self._original_json = json.dumps(self._original_typedef, indent=4)
        self._type_editor = burp_callbacks.createTextEditor()
        self._type_editor.setEditable(True)
        self._type_editor.setText(self._original_json)

        splitPane = JSplitPane(JSplitPane.HORIZONTAL_SPLIT)
        splitPane.setLeftComponent(self._type_editor.getComponent())
        splitPane.setRightComponent(self.createButtonPane())
        splitPane.setResizeWeight(0.8)

        self.add(splitPane)

        self.is_open = True

    def createButtonPane(self):
        """Create a new button pane with the type editor window"""
        self._button_listener = TypeEditorButtonListener(self)

        panel = JPanel()
        panel.setLayout(BoxLayout(panel, BoxLayout.Y_AXIS))

        panel.add(Box.createRigidArea(Dimension(0, 5)))
        panel.add(
            self.createButton("Validate", "validate", "Check if typedef is valid")
        )
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(self.createButton("Save", "save", "Save the typedef"))
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(self.createButton("Reset", "reset", "Reset to original"))
        panel.add(Box.createRigidArea(Dimension(0, 3)))
        panel.add(self.createButton("Exit", "exit", "Close window and reset"))
        panel.add(Box.createRigidArea(Dimension(0, 3)))

        return panel

    def createButton(self, text, command, tooltip):
        """Generate a new button with a given text and command"""
        button = JButton(text)
        button.setAlignmentX(Component.CENTER_ALIGNMENT)
        button.setActionCommand(command)
        button.addActionListener(self._button_listener)
        button.setToolTipText(tooltip)
        return button

    def applyType(self):
        """Callback for the apply button. Validates the definition and calls
        the callback provided when opening the window
        """

        try:
            new_json = self._type_editor.getText().tostring()
            message_type = json.loads(new_json)
            if self._original_json == new_json and len(message_type) != 0:
                # Detect if no text was changed, but allow an empty type for
                # adding empty ones
                self.exitTypeWindow()
                return

            blackboxprotobuf.validate_typedef(message_type, self._original_typedef)

            self._type_callback(message_type, self._type_source)
            self.exitTypeWindow()

        except Exception as exc:
            self._burp_callbacks.printError(traceback.format_exc())
            JOptionPane.showMessageDialog(self, "Error saving type: " + str(exc))

    def resetTypeWindow(self):
        """Callback for reset button. Resets to the original type definition"""
        self._type_editor.setText(json.dumps(self._original_typedef, indent=4))

    def exitTypeWindow(self):
        """Callback for exit button. Exits the window without saving"""
        self.is_open = False
        self.dispatchEvent(WindowEvent(self, WindowEvent.WINDOW_CLOSING))

    def validateType(self):
        """Callback for validate button. Validates the type without saving"""
        try:
            message_type = json.loads(self._type_editor.getText().tostring())
        except Exception as exc:
            self._burp_callbacks.printError(traceback.format_exc())
            JOptionPane.showMessageDialog(self, "Error decoding JSON: " + str(exc))
            return

        try:
            blackboxprotobuf.validate_typedef(message_type, self._original_typedef)
        except Exception as exc:
            self._burp_callbacks.printError(traceback.format_exc())
            JOptionPane.showMessageDialog(self, "Error validating type: " + str(exc))
            return


class TypeEditorButtonListener(ActionListener):
    """Button action listener for the type editor window"""

    def __init__(self, type_editor):
        self._type_editor = type_editor

    def actionPerformed(self, event):
        """Called when a button is pressed"""
        if event.getActionCommand() == "validate":
            self._type_editor.validateType()
        elif event.getActionCommand() == "save":
            self._type_editor.applyType()
        elif event.getActionCommand() == "reset":
            self._type_editor.resetTypeWindow()
        elif event.getActionCommand() == "exit":
            self._type_editor.exitTypeWindow()



"""Adds the top level type definition editor and registers the protobuf message
editor factory for individual tabs
"""

# Copyright (c) 2018-2023 NCC Group Plc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import inspect
import os
import sys
import traceback
import burp
import json
from javax.swing import DefaultListModel
from java.util.concurrent import Executors


# Add correct directory to sys.path
_BASE_DIR = os.path.abspath(
    os.path.dirname(inspect.getfile(inspect.currentframe())) + "../../../../"
)
sys.path.insert(0, _BASE_DIR + "/burp/")

import blackboxprotobuf
from blackboxprotobuf.lib.config import default as default_config
from blackboxprotobuf.burp import editor, typedef_tab, typedef_editor


EXTENSION_NAME = "BlackboxProtobuf"


class BurpExtender(burp.IBurpExtender, burp.IExtensionStateListener):
    """Primary extension class. Sets up all other functionality."""

    def __init__(self):
        self.open_windows = []
        self.callbacks = None
        self.helpers = None
        self.saved_types = {}
        self.suite_tab = None
        # global list of known messages for all views
        # this should mirror what is in blackboxprotobuf.known_messages
        # TODO bundle them together so it doesn't have to be manually updated
        self.known_message_model = DefaultListModel()
        self.refresh_message_model()

        self.thread_executor = Executors.newCachedThreadPool()

    def refresh_message_model(self):
        self.known_message_model.clear()
        for name in default_config.known_types.keys():
            self.known_message_model.addElement(name)

    def registerExtenderCallbacks(self, callbacks):
        """Called by burp. Collects callback object and sets up UI"""
        try:
            callbacks.registerExtensionStateListener(self)

            self.callbacks = callbacks

            self.helpers = callbacks.getHelpers()

            callbacks.setExtensionName(EXTENSION_NAME)

            callbacks.registerMessageEditorTabFactory(
                editor.ProtoBufEditorTabFactory(self, callbacks)
            )

            self.suite_tab = typedef_tab.TypeDefinitionTab(self, callbacks)
            callbacks.addSuiteTab(self.suite_tab)
            self.loadKnownMessages()
            self.refresh_message_model()
        except Exception as exc:
            self.callbacks.printError(traceback.format_exc())
            raise exc

    def loadKnownMessages(self):
        message_json = self.callbacks.loadExtensionSetting("known_types")
        if message_json:
            default_config.known_types.update(json.loads(message_json))
        saved_types = self.callbacks.loadExtensionSetting("saved_type_map")
        if saved_types:
            self.saved_types.update(json.loads(saved_types))

    def saveKnownMessages(self):
        # TODO might be good to cll this more often (eg. when messages are updated)
        # save the known messages
        self.callbacks.saveExtensionSetting(
            "known_types", json.dumps(default_config.known_types)
        )
        self.callbacks.saveExtensionSetting(
            "saved_type_map", json.dumps(self.saved_types)
        )

    def extensionUnloaded(self):
        self.thread_executor.shutdownNow()
        self.saveKnownMessages()
        for window in self.open_windows:
            window.exitTypeWindow()

    def open_typedef_editor(self, message_type, source, callback):
        self.open_windows = [window for window in self.open_windows if window.is_open]

        window = typedef_editor.TypeEditorWindow(
            self.callbacks,
            message_type,
            source,
            callback,
        )
        window.show()

        self.open_windows.append(window)
